from .storage import JsonStorage
from .scanner import Scanner
from .parser import ParserBase, Parser
from .util import load, DOC_INHERIT, state_size_dataset


class Markov(metaclass=DOC_INHERIT):
    """Markov chain generator base class.

    Attributes
    ----------
    DEFAULT_SCANNER : `type`
        Default scanner class.
    DEFAULT_PARSER : `type`
        Default parser class.
    DEFAULT_STORAGE : `type`
        Default storage class.
    scanner : `markovchain.scanner.Scanner`
    parser : `markovchain.parser.ParserBase`
    storage : `markovchain.storage.Storage`
    """
    DEFAULT_SCANNER = Scanner
    DEFAULT_PARSER = Parser
    DEFAULT_STORAGE = JsonStorage

    def __init__(self,
                 scanner=None,
                 parser=None,
                 storage=None):
        """Markov chain generator base class constructor.

        Parameters
        ----------
        scanner : `dict` or `markovchain.scanner.Scanner`, optional
            Scanner (default: `DEFAULT_SCANNER()`).
        parser : `dict` or `markovchain.parser.ParserBase`, optional
            Parser (default: `DEFAULT_PARSER()`).
        storage : `markovchain.storage.Storage`, optional
            Parser (default: `DEFAULT_STORAGE()`).
        """
        if storage is None:
            storage = self.DEFAULT_STORAGE()
        #if scanner is None:
        #    scanner = storage.settings.get('scanner', None)
        #if parser is None:
        #    scanner = storage.settings.get('parser', None)
        self.storage = storage
        self.scanner = load(scanner, Scanner, self.DEFAULT_SCANNER)
        self.parser = load(parser, ParserBase, self.DEFAULT_PARSER)

    def __eq__(self, markov):
        return (self.scanner == markov.scanner
                and self.parser == markov.parser
                and self.storage == markov.storage)

    def data(self, data, part=False, dataset=''):
        """Parse data and update links.

        Parameters
        ----------
        data
            Data to parse.
        part : `bool`, optional
            True if data is partial (default: `False`).
        dataset : `str`, optional
            Dataset key prefix (default: '').
        """
        links = self.parser(self.scanner(data, part), part, dataset)
        self.storage.add_links(links)

    def generate(self, state_size=None, start=(), dataset='', backward=False):
        """Generate a sequence.

        Parameters
        ----------
        state_size : `int`, optional
            State size (default: parser.state_sizes[0]).
        start : `str` or `iterable` of `str`, optional
            Initial state (default: ()).
        dataset : `str`, optional
            Dataset key prefix.
        backward : `bool`, optional
            Link direction.

        Returns
        -------
        `generator` of `str`
            State generator.
        """
        if state_size is None:
            try:
                state_size = next(iter(self.parser.state_sizes))
            except StopIteration:
                return
        #elif (self.parser is not None
        #      and state_size not in self.parser.state_sizes):
        #    raise ValueError('invalid state size: {0}: not in {1}'
        #                     .format(state_size, self.parser.state_sizes))
        dataset += state_size_dataset(state_size)
        return self.storage.generate(start, state_size, dataset, backward)

    def get_settings_json(self):
        """Convert generator settings to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        return {
            'scanner': None if self.scanner is None else self.scanner.save(),
            'parser': None if self.parser is None else self.parser.save()
        }

    def save(self, fp=None):
        """Save to file.

        Parameters
        ----------
        fp : `file`, optional
            Output file.
        """
        self.storage.settings['markov'] = self.get_settings_json()
        self.storage.save(fp)

    @classmethod
    def from_storage(cls, storage):
        """Load from storage.

        Parameters
        ----------
        storage : `markovchain.storage.Storage`

        Returns
        -------
        `markovchain.Markov`
        """
        args = dict(storage.settings.get('markov', {}))
        args['storage'] = storage
        return cls(**args)

    @classmethod
    def from_file(cls, fp, storage=None):
        """Load from file.

        Parameters
        ----------
        fp : `str` or `file`
            File or path.
        storage : `type`, optional
            Storage class (default: cls.DEFAULT_STORAGE)

        Returns
        -------
        `markovchain.Markov`
        """
        if storage is None:
            storage = cls.DEFAULT_STORAGE
        return cls.from_storage(storage.load(fp))

    @classmethod
    def from_settings(cls, settings=None, storage=None):
        """Create from settings.

        Parameters
        ----------
        settings : `dict`, optional
            Settings (default: None).
        storage : `type`, optional
            Storage class (default: cls.DEFAULT_STORAGE)

        Returns
        -------
        `markovchain.Markov`
        """
        if storage is None:
            storage = cls.DEFAULT_STORAGE
        return cls.from_storage(storage(settings=settings))



import io
import pytest


class MockCLI:
    class Exit(Exception):
        def __init__(self, status=0):
            super().__init__('Command exited with status %s' % status)
            self.status = status

    @classmethod
    def mock_exit(cls, status=0):
        raise cls.Exit(status)

    def __init__(self, tmpdir):
        self.dir = str(tmpdir)
        self.stdin = None
        self.stdout = None
        self.stderr = None
        self.exit = None

    def __call__(self, mocker):
        self.stdin = mocker.patch('sys.stdin', new_callable=io.StringIO)
        self.stdout = mocker.patch('sys.stdout', new_callable=io.StringIO)
        self.stderr = mocker.patch('sys.stderr', new_callable=io.StringIO)
        self.exit = mocker.patch('sys.exit', wraps=self.mock_exit)

    def run(self, cmd, *args, **kwargs):
        try:
            cmd(*args, **kwargs)
        except self.Exit:
            pass

    def reset(self):
        for fp in (self.stdin, self.stdout, self.stderr):
            fp.seek(0)
            fp.truncate()
        self.exit.reset_mock()

    def assert_output(self, stdout, stderr=None, exit_status=None):
        if stdout is not None:
            assert self.stdout.getvalue() == stdout
        if stderr is not None:
            assert self.stderr.getvalue() == stderr
        if exit_status is None:
            assert self.exit.call_count == 0
        else:
            self.exit.assert_called_once_with(exit_status)


@pytest.fixture
def mock_cli(tmpdir):
    return MockCLI(tmpdir)



import re
import enum


RE_PUNCT = re.compile(r'^[^\w\s]+$')
RE_WORD = re.compile(r'\w+')

RE_FLAGS = 'AILMSUX'
RE_CUSTOM_FLAGS = 'O'


class ReplyMode(enum.IntEnum):
    """Text reply mode.
    """
    END = 0
    START = 1
    REPLY = 2


class CharCase(enum.IntEnum):
    """Character case."""
    PRESERVE = 0
    TITLE = 1
    UPPER = 2
    LOWER = 3

    def convert(self, string):
        """Return a copy of string converted to case.

        Parameters
        ----------
        string : `str`

        Returns
        -------
        `str`

        Examples
        --------
        >>> CharCase.LOWER.convert('sTr InG')
        'str ing'
        >>> CharCase.UPPER.convert('sTr InG')
        'STR ING'
        >>> CharCase.TITLE.convert('sTr InG')
        'Str ing'
        >>> CharCase.PRESERVE.convert('sTr InG')
        'sTr InG'
        """
        if self == self.__class__.TITLE:
            return capitalize(string)
        if self == self.__class__.UPPER:
            return string.upper()
        if self == self.__class__.LOWER:
            return string.lower()
        return string


class ReFlags(enum.IntEnum):
    """Custom regexp flags.

    Attributes
    ----------
    O : `int`
    OVERLAP : `int`
        Replace overlapping occurrences of pattern.
    """
    O = 1
    OVERLAP = 1


def ispunct(string):
    """Return `True` if all characters in a string are punctuation
    and it is not empty.

    Parameters
    ----------
    string : `str`

    Returns
    -------
    `bool`

    Examples
    --------
    >>> ispunct('.,?')
    True
    >>> ispunct('.x.')
    False
    >>> ispunct('. ')
    False
    >>> ispunct('')
    False
    """
    return RE_PUNCT.match(string) is not None

def get_words(string):
    """Find all words in a string.

    Parameters
    ----------
    string : `str`

    Returns
    -------
    `list` of `str`

    Examples
    --------
    >>> get_words('  ..?!word  ,  (Word)..  word')
    ['word', 'Word', 'word']
    """
    return RE_WORD.findall(string)

def lstrip_ws_and_chars(string, chars):
    """Remove leading whitespace and characters from a string.

    Parameters
    ----------
    string : `str`
        String to strip.
    chars : `str`
        Characters to remove.

    Returns
    -------
    `str`
        Stripped string.

    Examples
    --------
    >>> lstrip_ws_and_chars(' \\t.\\n , .x. ', '.,?!')
    'x. '
    """
    res = string.lstrip().lstrip(chars)
    while len(res) != len(string):
        string = res
        res = string.lstrip().lstrip(chars)
    return res

def capitalize(string):
    """Capitalize a sentence.

    Parameters
    ----------
    string : `str`
        String to capitalize.

    Returns
    -------
    `str`
        Capitalized string.

    Examples
    --------
    >>> capitalize('worD WORD WoRd')
    'Word word word'
    """
    if not string:
        return string
    if len(string) == 1:
        return string.upper()
    return string[0].upper() + string[1:].lower()


def re_flags(flags, custom=ReFlags):
    """Parse regexp flag string.

    Parameters
    ----------
    flags: `str`
        Flag string.
    custom: `IntEnum`, optional
        Custom flag enum (default: None).

    Returns
    -------
    (`int`, `int`)
        (flags for `re.compile`, custom flags)

    Raises
    ------
    ValueError
    """
    re_, custom_ = 0, 0
    for flag in flags.upper():
        try:
            re_ |= getattr(re, flag)
        except AttributeError:
            if custom is not None:
                try:
                    custom_ |= getattr(custom, flag)
                except AttributeError:
                    raise ValueError('Invalid custom flag "%s"' % flag)
            else:
                raise ValueError('Invalid regexp flag "%s"' % flag)
    return re_, custom_

def re_flags_str(flags, custom_flags):
    """Convert regexp flags to string.

    Parameters
    ----------
    flags : `int`
        Flags.
    custom_flags : `int`
        Custom flags.

    Returns
    -------
    `str`
        Flag string.
    """
    res = ''
    for flag in RE_FLAGS:
        if flags & getattr(re, flag):
            res += flag
    for flag in RE_CUSTOM_FLAGS:
        if custom_flags & getattr(ReFlags, flag):
            res += flag
    return res

def re_sub(pattern, repl, string, count=0, flags=0, custom_flags=0):
    """Replace regular expression.

    Parameters
    ----------
    pattern : `str` or `_sre.SRE_Pattern`
        Compiled regular expression.
    repl : `str` or `function`
        Replacement.
    string : `str`
        Input string.
    count: `int`
        Maximum number of pattern occurrences.
    flags : `int`
        Flags.
    custom_flags : `int`
        Custom flags.
    """
    if custom_flags & ReFlags.OVERLAP:
        prev_string = None
        while string != prev_string:
            prev_string = string
            string = re.sub(pattern, repl, string, count, flags)
        return string
    return re.sub(pattern, repl, string, count, flags)



import pytest

from markovchain.scanner import Scanner
from markovchain.text.scanner import CharScanner, RegExpScanner


def scan_str(scanner, data, part=False, sep=''):
    return sep.join(word for word in scanner(data, part) if word != scanner.END)


@pytest.mark.parametrize('test,res', [
    (('', True), []),
    (('', False), []),
    (('ab c.d', True), list('ab c.d')),
    (('ab c.d', False), list('ab c.d') + [Scanner.END])
])
def test_char_scanner_id(test, res):
    scan = CharScanner(None, None)
    assert list(scan(*test)) == res

def test_char_scanner_default():
    scan = CharScanner()

    assert list(scan('ab..c')) == [
        'a', 'b', '.', '.', scan.END, 'c', '.', scan.END
    ]
    assert list(scan('a b..c', True)) == [
        'a', ' ', 'b', '.', '.', scan.END, 'c'
    ]
    assert list(scan('.', True)) == ['.']
    assert list(scan('', False)) == [scan.END]

    assert list(scan('abc', True)) == ['a', 'b', 'c']
    assert list(scan('', False)) == ['.', scan.END]

    assert list(scan('...')) == []

@pytest.mark.parametrize('test,res', [
    (('abc.de?!f',), 'abc.de?!f.'),
    (('.?!.a',), 'a.'),
    (('.?!.a', True), 'a'),
    (('a.',), 'a.')
])
def test_char_scanner_default_str(test, res):
    scan = CharScanner()
    assert scan_str(scan, *test) == res

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    ((None, None), (None, None), True),
    (('x', 'y'), ('x', 'y'), True),
    (('x', 'y'), ('x', 'y', 'upper'), False),
    (('x', 'y'), ('x', 'x'), False),
    (('x', 'y'), ('y', 'y'), False)
])
def test_char_scanner_eq(test, test2, res):
    scan = CharScanner(*test)
    scan2 = CharScanner(*test2)
    assert (scan == scan2) == res

@pytest.mark.parametrize('test', [
    (), (None, None), ('ab', 'cd', 'upper')
])
def test_char_scanner_save_load(test):
    scanner = CharScanner(*test)
    saved = scanner.save()
    loaded = Scanner.load(saved)
    assert scanner == loaded


@pytest.mark.parametrize('test,res', [
    (('', True), []),
    (('', False), []),
    (('ab c.d', True), list('ab c.d')),
    (('ab c.d', False), list('ab c.d') + [Scanner.END])
])
def test_regexp_scanner_id(test, res):
    scan = RegExpScanner('.', None)
    assert list(scan(*test)) == res

def test_regexp_scanner_default():
    scan = RegExpScanner()

    assert list(scan('ab..c')) == [
        'ab', '..', scan.END, 'c', '.', scan.END
    ]
    assert list(scan('a \n b?!. .. !! ??c', True)) == [
        'a', 'b', '?!.', scan.END, 'c'
    ]
    assert list(scan('.', True)) == ['.', scan.END]
    assert list(scan('', False)) == []
    assert list(scan('... .. . ! \n ? ?!  ')) == []

@pytest.mark.parametrize('test,res', [
    (('a\t\nb\nc.d   e ?!f',), 'abc.de?!f.'),
    (('.?!.a', True), 'a'),
    (('a.',), 'a.')
])
def test_regexp_scanner_default_str(test, res):
    scan = RegExpScanner()
    assert scan_str(scan, *test) == res

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    (('.', None), ('.', None), True),
    (('.', None), ('.', '.'), False),
    (('.', 'x'), ('.', 'x'), True),
    (('.', 'x'), ('.', 'x', 'upper'), False),
    (('x', '.'), ('.', '.'), False),
    (('x', '.'), ('x', 'x'), False)
])
def test_regexp_scanner_eq(test, test2, res):
    scan = RegExpScanner(*test)
    scan2 = RegExpScanner(*test2)
    assert (scan == scan2) == res

@pytest.mark.parametrize('test', [
    (), ('.*', ',', 'upper')
])
def test_regexp_scanner_save_load(test):
    scanner = RegExpScanner(*test)
    saved = scanner.save()
    loaded = Scanner.load(saved)
    assert scanner == loaded



from abc import abstractmethod
from random import randint

from ..util import DOC_INHERIT_ABSTRACT


class Storage(metaclass=DOC_INHERIT_ABSTRACT):
    """Storage base class.

    Attributes
    ----------
    settings : `dict`
    state_separator : `str`
    """

    def __init__(self, settings=None):
        """Storage base constructor.

        Parameters
        ----------
        settings: `dict`, optional
        """
        if settings is None:
            settings = {}
        self.settings = settings
        sep = settings.get('storage', {}).get('state_separator', ' ')
        self._state_separator = sep

    def __eq__(self, storage):
        return self.settings == storage.settings

    @property
    def state_separator(self):
        """`str` : State separator.
        """
        return self._state_separator

    @state_separator.setter
    def state_separator(self, separator):
        if self._state_separator is not None:
            self.replace_state_separator(self._state_separator, separator)
        self._state_separator = separator

    def split_state(self, state):
        """Split state string.

        Parameters
        ----------
        state : `str`

        Returns
        -------
        `list` of `str`
        """
        if self.state_separator:
            return state.split(self.state_separator)
        return list(state)

    def join_state(self, state):
        """Join states.

        Parameters
        ----------
        state : `iterable` of `str`

        Returns
        -------
        `str`
        """
        return self.state_separator.join(state)

    def random_link(self, dataset, state, backward=False):
        """Get a random link.

        Parameters
        ----------
        dataset : `object`
            Dataset from `self.get_dataset()`.
        state : `object`
            Link source.
        backward : `bool`, optional
            Link direction.

        Raises
        ------
        ValueError
            If link count is invalid.

        Returns
        -------
        (`str` or `None`, `object` or `None`)
            Link value and next state.
        """
        links = self.get_links(dataset, state, backward)
        if not links:
            return None, None
        x = randint(0, sum(link[0] for link in links) - 1)
        for link in links:
            count = link[0]
            if x < count:
                return link[1], self.follow_link(link, state, backward)
            x -= count
        raise RuntimeError('invalid link sum')

    def generate(self, state, size, dataset, backward=False):
        """Generate a sequence.

        Parameters
        ----------
        state : `str` or `iterable` of `str`
            Initial state.
        size : `int`
            State size.
        dataset : `str`
            Dataset key.
        backward : `bool`, optional
            Link direction.

        Returns
        -------
        `generator` of `str`
            Node value generator.
        """
        if isinstance(state, str):
            state = self.split_state(state)
        state = self.get_state(state, size)
        dataset = self.get_dataset(dataset)
        while True:
            link, state = self.random_link(dataset, state, backward)
            if link is None or backward and link == '':
                return
            yield link

    def save(self, fp=None):
        """Update settings JSON data and save to file.

        Parameters
        ----------
        fp : `file` or `str`, optional
            Output file.
        """
        self.settings['storage'] = {
            'state_separator': self.state_separator
        }
        self.do_save(fp)

    @abstractmethod
    def get_dataset(self, key, create=False):
        """Get data set by key.

        Parameters
        ----------
        key : `str`
            Dataset key.
        create : `bool`, optional
            Create dataset if it does not exist.

        Returns
        -------
        `object`
            Dataset.

        Raises
        ------
        KeyError
            If dataset does not exist and `create` == `False`.
        """
        pass

    @abstractmethod
    def replace_state_separator(self, old_separator, new_separator):
        """Replace state separator.

        Parameters
        ----------
        old_separator : `str`
            Old state separator.
        new_separator : `str`
            New state separator.
        """
        pass

    @abstractmethod
    def add_links(self, links, dataset_prefix=''):
        """Add links.

        Parameters
        ----------
        links : `generator` of (`str`, `islice` of `str`, `str`)
            Links to add.
        dataset_prefix : `str`, optional
            Dataset key prefix.
        """
        pass

    @abstractmethod
    def get_state(self, state, size):
        """Convert strings to state.

        Parameters
        ----------
        state : `iterable` of `str`
            Node values.
        size : `int`
            State size.

        Returns
        -------
        `object`
            State.
        """
        pass

    @abstractmethod
    def get_states(self, dataset, string):
        """Get all states containing a substring.

        Parameters
        ----------
        dataset : `str`
            Dataset key.
        string : `str`
            String to search.

        Returns
        -------
        `list` of `str`
            States.
        """
        pass

    @abstractmethod
    def get_links(self, dataset, state, backward=False):
        """Get links.

        Parameters
        ----------
        dataset : `object`
            Dataset from `self.get_dataset()`.
        state : `object`
            State from `self.get_state()`.
        backward : `bool`, optional
            Link direction.

        Returns
        -------
        `None` or `list` of (`int`, `str`, `object`)
            Links (count, value, data).
        """
        pass

    @abstractmethod
    def follow_link(self, link, state, backward=False):
        """Follow a link.

        Parameters
        ----------
        link : (`int`, `str`, `object`)
            Link to follow.
        state : `object`
            State.
        backward : `bool`, optional
            Link direction.

        Returns
        -------
        `object`
            New state.
        """
        pass

    @abstractmethod
    def do_save(self, fp=None):
        """Save to file.

        Parameters
        ----------
        fp : `file` or `str`, optional
            Output file.
        """
        pass

    @classmethod
    @abstractmethod
    def load(cls, fp):
        """Load from file.

        Parameters
        ----------
        fp : `file` or `str`, optional
            Input file or path.

        Returns
        -------
        `markovchain.storage.Storage`
            Loaded storage.
        """
        pass



from collections import deque
from unittest.mock import Mock, call
import pytest

from markovchain import Markov
from markovchain.parser import ParserBase
from markovchain.scanner import Scanner


def test_markov_base_properties():
    storage = Mock(settings={})
    markov = Markov(storage=storage)
    assert isinstance(markov.parser, ParserBase)
    assert isinstance(markov.scanner, Scanner)
    assert markov.storage is storage

@pytest.mark.parametrize('data,part,dataset', [
    ([], False, ''),
    ([1, 2, 3], True, 'data')
])
def test_markov_base_data(data, part, dataset):
    storage = Mock(settings={})
    scanner = Mock(return_value=0)
    parser = Mock(return_value=1)
    markov = Markov(parser=parser, scanner=scanner, storage=storage)
    markov.data(data, part, dataset)
    scanner.assert_called_once_with(data, part)
    parser.assert_called_once_with(0, part, dataset)
    storage.add_links.assert_called_once_with(1)

@pytest.mark.parametrize('state_sizes, args, call', [
    ([1], (1, 'x', 'd'), ('x', 1, 'd_1', False)),
    ([2, 1], (None, 'y', 'dd', True), ('y', 2, 'dd_2', True)),
    ([], (None, 'y', 'dd'), None)
])
def test_markov_base_generate(mocker, state_sizes, args, call):
    mocker.patch(
        'markovchain.base.state_size_dataset',
        wraps=lambda ss: '_%d' % ss
    )
    markov = Markov(
        parser=Mock(state_sizes=state_sizes),
        scanner=Mock(),
        storage=Mock(generate=Mock(return_value=0))
    )
    res = markov.generate(*args)
    if call is not None:
        assert res == 0
        markov.storage.generate.assert_called_with(*call)
    else:
        assert res is None
        assert markov.storage.generate.call_count == 0

def test_markov_base_get_settings_json():
    markov = Markov(
        parser=Mock(save=lambda: 0),
        scanner=Mock(save=lambda: 1),
        storage=Mock()
    )
    json = markov.get_settings_json()
    assert json == {
        'scanner': 1,
        'parser': 0
    }

@pytest.mark.parametrize('test,res', [
    (((0, 0, 0), (0, 0, 0)), True),
    (((0, 0, 0), (0, 0, 1)), False),
    (((0, 0, 0), (0, 1, 0)), False),
    (((0, 0, 0), (1, 0, 0)), False)
])
def test_markov_eq(test, res):
    markov = []
    for scanner, parser, storage in test:
        m = Markov()
        m.scanner = scanner
        m.parser = parser
        m.storage = storage
        markov.append(m)
    assert (markov[0] == markov[1]) == res

def test_markov_base_save(mocker):
    mocker.patch('markovchain.Markov.get_settings_json', return_value=0)
    storage = Mock(settings={})
    markov = Markov(
        parser=Mock(),
        scanner=Mock(),
        storage=storage
    )
    markov.save(1)
    assert storage.settings == {'markov': 0}
    storage.save.assert_called_once_with(1)
    markov.get_settings_json.assert_called_once_with() # pylint:disable=no-member



from itertools import product
import pytest

from markovchain.image.traversal import (
    Traversal, HLines, VLines, Spiral, Hilbert, Blocks
)


@pytest.mark.parametrize('args,test,res', [
    ((0,), (2, 3, False), [(0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2)]),
    ((0,), (2, 3, True), [(0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2)]),
    ((1,), (2, 3), [(1, 0), (0, 0), (1, 1), (0, 1), (1, 2), (0, 2)]),
    ((2,), (2, 3), [(1, 0), (0, 0), (0, 1), (1, 1), (1, 2), (0, 2)]),
    ((0, True), (2, 3, False),
     [(0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2)]),
    ((0, True), (2, 3, True),
     [(0, 0), (1, 0), None, (0, 1), (1, 1), None, (0, 2), (1, 2), None])
])
def test_hlines_traverse(args, test, res):
    assert list(HLines(*args)(*test)) == res

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    ((1, True), (1, True), True),
    ((0, False), (0, True), False),
    ((0, False), (1, False), False)
])
def test_hlines_eq(test, test2, res):
    assert (HLines(*test) == HLines(*test2)) == res

@pytest.mark.parametrize('test', [
    (0, False), (2, True)
])
def test_hlines_save_load(test):
    test = HLines(*test)
    saved = test.save()
    loaded = Traversal.load(saved)
    assert isinstance(loaded, HLines)
    assert test == loaded


@pytest.mark.parametrize('args,test,res', [
    ((0,), (2, 3, False), [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]),
    ((0,), (2, 3, True), [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]),
    ((1,), (2, 3), [(0, 2), (0, 1), (0, 0), (1, 2), (1, 1), (1, 0)]),
    ((2,), (2, 3), [(0, 2), (0, 1), (0, 0), (1, 0), (1, 1), (1, 2)]),
    ((0, True), (2, 3, False),
     [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]),
    ((0, True), (2, 3, True),
     [(0, 0), (0, 1), (0, 2), None, (1, 0), (1, 1), (1, 2), None])
])
def test_vlines_traverse(args, test, res):
    assert list(VLines(*args)(*test)) == res

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    ((1, True), (1, True), True),
    ((0, False), (0, True), False),
    ((0, False), (1, False), False)
])
def test_vlines_eq(test, test2, res):
    assert (VLines(*test) == VLines(*test2)) == res

@pytest.mark.parametrize('test', [
    (0, False), (2, True)
])
def test_vlines_save_load(test):
    test = VLines(*test)
    saved = test.save()
    loaded = Traversal.load(saved)
    assert isinstance(loaded, VLines)
    assert test == loaded


@pytest.mark.parametrize('width,height', product(range(1, 7), range(1, 7)))
def test_spiral_traverse(width, height):
    test = Spiral()
    test.reverse = False
    res = list(test(width, height, False))
    assert sorted(res) == list(product(range(width), range(height)))
    test.reverse = True
    res2 = list(test(width, height, False))
    res.reverse()
    assert res2 == res

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    ((True,), (False,), False),
    ((True,), (True,), True)
])
def test_spiral_eq(test, test2, res):
    assert (Spiral(*test) == Spiral(*test2)) == res

@pytest.mark.parametrize('test', [
    (False,), (True,)
])
def test_spiral_save_load(test):
    test = Spiral(*test)
    saved = test.save()
    loaded = Traversal.load(saved)
    assert isinstance(loaded, Spiral)
    assert test == loaded


@pytest.mark.parametrize('width,height', product(range(1, 7), range(1, 7)))
def test_hilbert_traverse(width, height):
    test = Hilbert()
    res = list(test(width, height))
    assert sorted(res) == list(product(range(width), range(height)))

def test_hilbert_save_load():
    test = Hilbert()
    saved = test.save()
    loaded = Traversal.load(saved)
    assert isinstance(loaded, Hilbert)
    assert test == loaded


@pytest.mark.parametrize('test,res,res2', [
    (
        (4, 4, False),
        [(0, 0), (0, 1), (2, 0), (2, 1)],
        [(0, 0), (0, 1), (2, 0), (2, 1)]
    ),
    (
        (3, 3, True),
        [(0, 0), (0, 1), None],
        [(0, 0), (0, 1), None, None]
    ),
    (
        (4, 4, True),
        [(0, 0), (0, 1), None, (2, 0), (2, 1)],
        [(0, 0), (0, 1), None, None, (2, 0), (2, 1), None]
    )
])
def test_blocks_traverse(test, res, res2):
    def hline(width, height, ends): # pylint: disable=unused-argument
        for x in range(width):
            yield (x, 0)
            if ends and x == 0:
                yield None

    def vline(width, height, ends): # pylint: disable=unused-argument
        for y in range(height):
            yield (0, y)
            if ends:
                yield None

    traverse = Blocks(block_size=(2, 2),
                      block_sentences=False,
                      traverse_image=hline,
                      traverse_block=vline)
    assert list(traverse(*test)) == res
    traverse.block_sentences = True
    assert list(traverse(*test)) == res2

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    (((8, 8),), ((4, 4),), False),
    (((4, 4), True), ((4, 4), False), False),
    (((4, 4), True, 0), ((4, 4), True, 1), False),
    (((4, 4), True, 0, 0), ((4, 4), True, 0, 1), False)
])
def test_blocks_eq(test, test2, res):
    assert (Blocks(*test) == Blocks(*test2)) == res

@pytest.mark.parametrize('test', [
    ((8, 8), True,
     HLines(reverse=1, line_sentences=True),
     VLines(reverse=2, line_sentences=False))
])
def test_blocks_save_load(test):
    test = Blocks(*test)
    saved = test.save()
    loaded = Traversal.load(saved)
    assert isinstance(loaded, Blocks)
    assert test == loaded



import sys
from argparse import ArgumentParser

from . import text
from .util import set_args
from ..info import CLI_VERSION

try:
    from . import image
except ImportError:
    image = None

def main(args=None):
    """CLI main function.

    Parameters
    ----------
    args : `list` of `str`, optional
        CLI arguments (default: `sys.argv`).
    """
    parser = ArgumentParser()
    parser.add_argument('-v', '--version',
                        action='version', version=CLI_VERSION)

    parsers = parser.add_subparsers(dest='dtype')

    text.create_arg_parser(parsers.add_parser('text'))
    if image is not None:
        image.create_arg_parser(parsers.add_parser('image'))

    if len(sys.argv if args is None else args) <= 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args)
    dtype = globals()[args.dtype]
    try:
        set_args(args)
        cmd = getattr(dtype, 'cmd_' + args.command)
        cmd(args)
    except ValueError as err:
        print(str(err), file=sys.stderr)
        sys.exit(1)



from itertools import repeat, islice

from .scanner import ImageScanner
from .type import ImageType, Indexed
from ..base import Markov
from ..parser import LevelParser
from ..util import load, fill, level_dataset
from .util import pixel_to_state, state_to_pixel


class MarkovImage(Markov):
    """Markov image generator.
    """

    DEFAULT_SCANNER = ImageScanner
    DEFAULT_PARSER = LevelParser
    DEFAULT_IMGTYPE = Indexed

    def __init__(self,
                 levels=1,
                 imgtype=None,
                 *args, **kwargs):
        """Markov image generator constructor.

        Parameters
        ----------
        levels : `int`, optional
            Number of levels.
        imgtype : `None` or `dict` or `markovchain.image.type.ImageType`
        """
        super().__init__(*args, **kwargs)
        self.imgtype = load(imgtype, ImageType, self.DEFAULT_IMGTYPE)
        self._levels = None
        self.levels = levels

    @property
    def levels(self):
        """`int` : Number of levels.
        """
        return self._levels

    @levels.setter
    def levels(self, levels):
        if self.scanner is not None:
            self.scanner.levels = levels
        if self.parser is not None:
            self.parser.levels = levels
        self._levels = levels

    def data(self, data, part=False, dataset=''):
        """
        Parameters
        ----------
        data : `PIL.Image`
            Image to parse.
        part : `bool`, optional
            True if data is partial (default: `False`).
        dataset : `str`, optional
            Dataset key prefix (default: '').
        """
        #if self.parser is None:
        #    raise ValueError('no parser')
        imgs = self.imgtype.convert(data)
        for channel, data in zip(self.imgtype.channels, imgs):
            key = dataset + channel
            data = self.scanner(data, part)
            if isinstance(self.parser, LevelParser):
                self.storage.add_links(self.parser(data, part, key))
            else:
                for level, level_data in enumerate(data):
                    level_key = key + level_dataset(level)
                    level_data = self.parser(level_data, part, level_key)
                    self.storage.add_links(level_data)

    def get_settings_json(self):
        data = super().get_settings_json()
        data['imgtype'] = self.imgtype.save()
        data['levels'] = self.levels
        return data

    def __eq__(self, markov):
        return (self.imgtype == markov.imgtype
                and self.levels == markov.levels
                and super().__eq__(markov))

    def __call__(self, width, height,
                 state_size=None, levels=None,
                 start_level=-1, start_image=None,
                 dataset=''):
        """Generate an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        state_size : `None` or `int` or `list` of `int`, optional
            State size (default: `None`).
        levels : `int`, optional
            Number of levels to generate (default: `self.scanner.levels`).
        start_level : `int`, optional
            Initial level (default: -1).
        start_image : `PIL.Image` or `None`
            Initial level image (default: `None`).
        dataset : `str`, optional
            Dataset key prefix (default: '').

        Returns
        -------
        `PIL.Image`
            Generated image.

        Raises
        ------
        ValueError
        """
        if start_level is None:
            if start_image is not None:
                start_level = 0
        elif start_level < 0:
            start_image = None
        elif start_level + 1 >= self.levels:
            if start_image is None:
                raise ValueError('invalid start level: {0}'.format(start_level))
            return start_image

        if start_image is not None:
            width, height = start_image.size
            start_image = self.imgtype.convert(start_image)
        else:
            start_image = repeat(None)
            start_level = -1

        if levels is None:
            levels = self.scanner.levels - start_level - 1
        elif levels <= 0 or start_level + levels >= self.levels:
            raise ValueError('invalid level count: {0}'.format(levels))

        state_sizes = fill(state_size, levels)
        for i, state_size in enumerate(state_sizes):
            if state_size is None:
                if isinstance(self.parser, LevelParser):
                    level = start_level + 1 + i
                    state_sizes[i] = self.parser.parsers[level].state_sizes[0]
                else:
                    state_sizes[i] = self.parser.state_sizes[0]

        channels = [
            self._channel(
                width, height, state_sizes,
                start_level, img, dataset + channel
            )
            for channel, img in zip(self.imgtype.channels, start_image)
        ]

        return self.imgtype.merge(channels)

    def _imgdata(self, width, height,
                 state_size=None, start='', dataset=''):
        """Generate image pixels.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        state_size : `int` or `None`, optional
            State size to use for generation (default: `None`).
        start : `str`, optional
            Initial state (default: '').
        dataset : `str`, optional
            Dataset key prefix (default: '').

        Raises
        ------
        RuntimeError
            If generator is empty.

        Returns
        -------
        `generator` of `int`
            Pixel generator.
        """
        size = width * height
        if size > 0 and start:
            yield state_to_pixel(start)
            size -= 1

        while size > 0:
            prev_size = size

            pixels = self.generate(state_size, start, dataset)
            pixels = islice(pixels, 0, size)

            for pixel in pixels:
                yield state_to_pixel(pixel)
                size -= 1

            if prev_size == size:
                if start:
                    yield from repeat(state_to_pixel(start), size)
                else:
                    raise RuntimeError('empty generator')

    @staticmethod
    def _write_imgdata(img, data, tr, x=0, y=0):
        """Write image data.

        Parameters
        ----------
        img : `PIL.Image.Image`
            Image.
        data : `iterable` of `int`
            Image data.
        tr : `markovchain.image.traversal.Traversal`
            Image traversal.
        x : `int`
            X offset.
        y : `int`
            Y offset.
        """
        for pixel, (x1, y1) in zip(data, tr):
            img.putpixel((x + x1, y + y1), pixel)
        return img

    def _channel(self, width, height, state_sizes,
                 start_level, start_image, dataset):
        """Generate a channel.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        state_sizes : `list` of (`int` or `None`)
            Level state sizes.
        start_level : `int`
            Initial level.
        start_image : `PIL.Image` or `None`
            Initial level image.
        dataset : `str`
            Dataset key prefix.

        Returns
        -------
        `PIL.Image`
            Generated image.
        """
        ret = start_image
        for level, state_size in enumerate(state_sizes, start_level + 1):
            key = dataset + level_dataset(level)
            if start_image is not None:
                scale = self.scanner.level_scale[level - 1]
                width *= scale
                height *= scale
            ret = self.imgtype.create_channel(width, height)
            if start_image is None:
                tr = self.scanner.traversal[0](width, height, ends=False)
                data = self._imgdata(width, height, state_size, '', key)
                self._write_imgdata(ret, data, tr)
            else:
                tr = self.scanner.traversal[0](
                    start_image.size[0],
                    start_image.size[1],
                    ends=False
                )
                for xy in tr:
                    start = pixel_to_state(start_image.getpixel(xy))
                    data = self._imgdata(scale, scale, state_size, start, key)
                    blk = self.scanner.traversal[level](scale, scale, False)
                    x, y = xy
                    x *= scale
                    y *= scale
                    self._write_imgdata(ret, data, blk, x, y)
            start_image = ret
        return ret



from unittest.mock import Mock
import pytest
from PIL import Image

from markovchain import Scanner, Parser, LevelParser
from markovchain.image import MarkovImage, ImageScanner, HLines, VLines


def test_markov_image_properties():
    markov = MarkovImage()
    assert isinstance(markov.scanner, ImageScanner)
    assert isinstance(markov.parser, LevelParser)
    assert markov.scanner.levels == markov.levels
    assert markov.parser.levels == markov.levels
    markov.levels = 3
    assert markov.scanner.levels == markov.levels
    assert markov.parser.levels == markov.levels
    with pytest.raises(ValueError):
        markov.levels = -1
    assert markov.scanner.levels == markov.levels
    assert markov.parser.levels == markov.levels
    assert markov.levels == 3

def test_markov_image_generate_error():
    markov = MarkovImage()
    with pytest.raises(KeyError):
        markov(2, 2)

@pytest.mark.parametrize('test,res', [
    ((2, 2), [0, 1, 2, 3]),
    ((4, 2), [0, 1, 2, 3, 0, 1, 2, 3])
])
def test_markov_image_generate(test, res):
    scanner = Scanner(lambda x: x)
    scanner.traversal = [HLines()]
    scanner.levels = 1
    scanner.level_scale = []
    markov = MarkovImage(
        scanner=scanner,
        parser=Parser()
    )
    markov.imgtype.convert = lambda x: [x]
    markov.data([['\x00', '\x01', '\x02', '\x03']])
    assert list(markov(*test).getdata()) == res

@pytest.mark.parametrize('args,kwargs,data,res', [
    ((2, 2), {}, False, KeyError),
    ((2, 2), {'levels': 0}, True, ValueError),
    ((2, 2), {'levels': 3}, True, ValueError),
    ((2, 2), {'levels': 1}, True, [0, 1, 2, 3]),
    (
        (2, 2),
        {},
        True,
        [
            0, 2, 1, 3,
            1, 3, 2, 0,
            2, 0, 3, 1,
            3, 1, 0, 2
        ]
    ),
    (
        (2, 1),
        {'levels': 1, 'start_level': -5, 'start_image': True},
        True,
        [0, 1]
    ),
    (
        (None, None),
        {'levels': 1, 'start_level': 0, 'start_image': True},
        True,
        [1, 3, 2, 0]
    ),
    (
        (None, None),
        {'levels': 1, 'start_level': 2, 'start_image': True},
        True,
        [1]
    ),
    (
        (None, None),
        {'levels': 2, 'start_level': 0, 'start_image': True},
        True,
        ValueError
    ),
])
def test_markov_image_generate_levels(args, kwargs, data, res):
    scanner = Scanner(lambda x: x)
    scanner.traversal = [HLines(), VLines()]
    scanner.levels = 2
    scanner.level_scale = [2]

    markov = MarkovImage(levels=2, scanner=scanner)
    markov.imgtype.convert = lambda x: [x]

    if data:
        markov.data([
            ['\x00', '\x01', '\x02', '\x03'],
            [(Scanner.START, '\x00'), '\x01',
             (Scanner.START, '\x01'), '\x02',
             (Scanner.START, '\x02'), '\x03',
             (Scanner.START, '\x03'), '\x00']
        ])

    if 'start_image' in kwargs:
        img = Image.new('P', (1, 1))
        img.putpixel((0, 0), 1)
        kwargs['start_image'] = img

    if isinstance(res, type):
        with pytest.raises(res):
            markov(*args, **kwargs)
    else:
        assert list(markov(*args, **kwargs).getdata()) == res

def test_markov_image_get_settings_json(mocker):
    get_settings_json = mocker.patch(
        'markovchain.Markov.get_settings_json',
        return_value={'x': 0}
    )
    markov = MarkovImage(
        levels=2,
        imgtype=Mock(save=lambda: 3)
    )
    data = markov.get_settings_json()
    assert data == {
        'x': 0,
        'levels': 2,
        'imgtype': 3
    }
    get_settings_json.assert_called_once_with()

@pytest.mark.parametrize('test,test2,res', [
    ((), (), True),
    ((1,), (2,), False),
    ((1, [8, 4, 8]), (1, [8, 4, 4]), False)
])
def test_markov_image_eq(test, test2, res):
    assert (MarkovImage(*test) == MarkovImage(*test2)) == res
